

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cmath>
#include <vector>
#include <map>

#include "plink.h"
#include "options.h"
#include "phase.h"
#include "helper.h"

extern ofstream LOG;

using namespace std;

// Format for haplotype file: format 1:
// SNP_ID  CHR  CM  BP  A1 A2   HAP   SNP1  SNP2 ... 

// i.e. length of Pred_allele indicate how many SNPs to expect:
// rs10001 5 0 10203  A G   TTC  rs001 rs002 rs003

// Alternatively: using wild cards: format 2:
// will name haplotype "H1_TTC_", "H1_CTT_", "H2_AA_", etc
// * rs001 rs002 rs003
// * rs002 rs004

void Plink::readTagFile()
{
 
  ///////////////////////////////////////////////
  // Haplotype inference is a SNP-major function
  
  if (!par::SNP_major) Ind2SNP();


  ////////////////////////
  // Lookup table for SNPs

  map<string,int> mlocus;
  map<string,int>::iterator ilocus;
  
  for (int l=0;l<nl_all;l++)
    mlocus.insert(make_pair(locus[l]->name,l));


  ///////////////////
  // Affection coding

  if (par::bt) affCoding(*this);

  
  ////////////////////////////////
  // Read list of tags/haplotypes

  checkFileExists(par::tagfile);
  ifstream TAG(par::tagfile.c_str(), ios::in);
  TAG.clear();
  
  string f2 = par::output_file_name + ".mishap";
  printLOG("Writing mis-specified haplotypes (if any) to [ " + f2 + " ]\n");
  ofstream MISHAP(f2.c_str(), ios::out);
  

  // Count of new haplotypes we want to infer
  int hc=1; 
  
  while(!TAG.eof())
    {
      char c[256];
      TAG.getline(c,256,'\n');
      
      string l = c;

      // Catch blank lines of DOS carriage-returns
      if (l=="" || l=="\r") continue;

      // Tokenize line
      string buf; 
      stringstream ss(l); 
      vector<string> tokens; 
      while (ss >> buf)
	tokens.push_back(buf);
      
      // whitepsace line?
      if (tokens.size() == 0)
	continue;
      
      // tokens[0]  predicted SNP rs#
      // tokens[1]  predicted SNP chromosome
      // tokens[2]  predicted SNP Morgan position
      // tokens[3]  predicted SNP base pair position
      // tokens[4]  predicted SNP allele 1
      // tokens[5]  predicted SNP allele 2
      // tokens[6]  tag allele (-> allele 1)
      // tokens[7+] predictor rs#(s)

      if (tokens.size() < 8 && tokens[0] != "*" ) 
	{
	  string e = "Problem with " + par::tagfile + " line\n" 
	    + l + "\n (expecting at least 8 items, or to start with * wildcard)\n";
	  error(e);	  
	}
      
      if (tokens[0].substr(tokens[0].size()-1) == "_" ) 
	error("Cannot use '_' in tag/haplotype name: reserved for wildcards\n");

      int len; 		// length of haplotype
      vector<int> tmp;	// list of predictor #s
      

      
      // Is this particular line a wildcard? okay?
      bool wildcard = tokens[0] == "*" ? true : false ;
      bool okay = true;


      /////////////////////////////////
      // Fully-specified haplotype 
      
      if (!wildcard) 
	{
	  len = tokens[6].length();
	  if (len != tokens.size() - 7)
	    {
	      string e = "Problem with " + par::tagfile + " line\n" + l + "\n";
	      error(e);
	    }      
	  
	  // Lookup locus name
	  for (int i=0; i<len; i++)
	    {
	      ilocus = mlocus.find(tokens[i+7]);
	      if (ilocus != mlocus.end())
		{
		  if (par::chr_sex[locus[ilocus->second]->chr] || 
		      par::chr_haploid[locus[ilocus->second]->chr]) 
		    {
		      MISHAP << "HAPLOID\t" << tokens[i+7] << "\n";
		      okay = false;
		    }
		  else
		    tmp.push_back(ilocus->second);	  	  
		}
	      else 
		{
		  MISHAP << "NOSNP\t" << tokens[0] 
			 << "\t" << tokens[i+7] << "\n";
		  okay = false;
		}
	    }
	}
      else
	{
	  /////////////////////////////////
	  // Wildcard selection
	  
          len = tokens.size()-1;

          // Lookup locus name
          for (int i=0; i<len; i++)
	    {
	      ilocus = mlocus.find(tokens[i+1]);
	      if (ilocus != mlocus.end())
		{
		  if (par::chr_sex[locus[ilocus->second]->chr] || 
		      par::chr_haploid[locus[ilocus->second]->chr]) 
		    {
		      MISHAP << "HAPLOID\t" << tokens[i+1] << "\n";
		      okay = false;
		    }
		  else
		    tmp.push_back(ilocus->second);	  	  
		}
	      else 
		{
		  MISHAP << "NOSNP\t" << tokens[i+1] << "\n";
		  okay = false;
		}
	    }
	}
      


      //////////////////////////////////////
      // Should we try to add this haploype

      if (!okay) continue;


      ////////////////////////////////////////////////////////
      // Check that all predictors are on the same chromosome
      
      if (!wildcard)  
	{    
	  for (int ck=0; ck<len; ck++)
	    if (locus[tmp[ck]]->chr != atoi(tokens[1].c_str()))
	      {
		MISHAP << "DIFF_CHR\t" << locus[tmp[ck]]->name 
		       << "\t" << tokens[0] << "\n";	  
		okay = false;
	      }
	}
      else
	{	
	  for (int ck=0; ck<len-1; ck++)
	    if (locus[tmp[ck]]->chr != locus[tmp[ck+1]]->chr)
	      {
		MISHAP << "DIFF_CHR\t" << locus[tmp[ck]]->name 
		       << "\t" << locus[tmp[ck+1]]->name << "\n";
		okay = false;
	      }
	}
      
      if (!okay) continue;
      
      ///////////////////////////////////////
      // Standard approach -- only one entry
      
      if (!wildcard)
	{
	  
	  // Add numbers for predictors to list
	  new_pred_locus.push_back(tmp);
	  
	  // Add which allele to look for (corresponding to allele1)
	  new_pred_allele.push_back(tokens[6]);
	  
	  // Make new entry in MAP file
	  Locus * loc = new Locus;
	  loc->name = tokens[0];
	  loc->chr = atoi(tokens[1].c_str()); // NOTE -- ADD X/Y CONVERSION
	  loc->pos = atof(tokens[2].c_str());
	  loc->bp = atoi(tokens[3].c_str());  
	  loc->allele1 = tokens[4];
	  loc->allele2 = tokens[5];
	  
	  // Add this new locus to the list
	  new_map.push_back(loc);
	}
      else
	{
	  
	  /////////////////////////////////////////
	  // Wildcard approach -- mutliple entries

	  HaploPhase H2(*this);
	  vector<string> poshap = H2.returnHaplotypes(tmp);
	  

	  // One entry per possible haplotype
	  for (int ss=0; ss<poshap.size(); ss++)
	    {
	      
	      // Add numbers for predictors to list
	      new_pred_locus.push_back(tmp);
	      
	      // Add which allele to look for (corresponding to allele1)
	      new_pred_allele.push_back(poshap[ss]);
		
	      // Make new entry in MAP file
		
	      Locus * loc = new Locus;
	      loc->name = "H"+int2str(hc)+"_"+poshap[ss]+"_";
	      loc->chr = locus[tmp[0]]->chr;
	      loc->pos = 0;
	      loc->bp = hc; 
	      loc->allele1 = "1";
	      loc->allele2 = "2";
	      
	      // Add this new locus to the list
	      new_map.push_back(loc);
	      
	    }
	  
	}
      
      // Increment new haplotype count
      hc++;
      
      // Read next in TAG file
      
    }

  TAG.close();
  MISHAP.close();
  
  // End of reading haplotype list -- did we encounter any problems?

  printLOG("Read " + int2str(new_map.size()) + 
	   " haplotypes from [ " + par::tagfile + " ]\n");  

  


  ///////////////////////////////////////////////
  // Impute missing SNPs -- create a new datafile
  
  // If any predictor is missing -> make imputed SNP missing If less
  // than par::impute_prob (default = 0.8) chance of most likely
  // phase, then impute missing
  
  // Fill in missing and homozygous calls first; use E-M algorithm to
  // tidy up the remaining SNPs
  
  
  // Place where new (common) SNPs will be deposited 
  
  vector<vector<bool> > new_one(n);
  vector<vector<bool> > new_two(n);
  
  for (int i=0; i<n; i++)
    {
      new_one[i].resize(0);
      new_two[i].resize(0);
    }


  ////////////////////////////////
  // For each new SNPs/haplotypes

  HaploPhase H(*this);

  stringstream s2;
  s2 << "\nAbout to estimate haplotype frequencies/phases ( MAF >= " 
     << par::min_af << " )\n";
  printLOG(s2.str());

  printLOG("Considering phases P(H|G) >= " +dbl2str(par::hap_min_phase_prob)+"\n");
  printLOG("Requiring per individual per haplotype genotyping of " +dbl2str(par::hap_missing_geno)+" \n");


  // Count number of founders
  
  int cnt_f = 0;
  vector<Individual*>::iterator person = sample.begin();
  while ( person != sample.end() ) 
    {  
      if ( (*person)->founder ) cnt_f++;	    
      person++;
    }
  if (cnt_f<n)
    {
      H.nonfounders = true;
      printLOG("Initial phasing based on "+
	       int2str(cnt_f)+" founders ("+
	       int2str(n-cnt_f)+
	       " non-founders)\n");
    }

  if (n == cnt_f && par::test_hap_TDT)
    error("Can not perform TDT in sample with no non-founders");

  // Weights (temporarily exclude this from analysis)
  ofstream WGT;
  if (false)
    {
      string f = par::output_file_name + ".weight";
      printLOG("Writing SNP-weights to [ " + f + " ]\n");
      WGT.open(f.c_str(), ios::out);
    }

  ofstream HFRQ;
  string f;

  if (par::display_hap_freqs) f = par::output_file_name + ".frq.hap";
  else if ( par::test_hap_CC ) f =  par::output_file_name + ".assoc.hap";
  else if ( par::test_hap_TDT ) f = par::output_file_name + ".tdt.hap";

  if (par::display_hap_freqs) 
    {
      printLOG("Writing haplotype frequencies to [ " + f + " ]\n");
      HFRQ.open(f.c_str(), ios::out);
      HFRQ.precision(4);

      
      HFRQ << setw(10) << "LOCUS" << " " 
	   << setw(12) << "HAPLOTYPE" << " "
	   << setw(10) << "F" << "\n";
      
    }
  
  if (par::test_hap_CC)
    {

      printLOG("Writing haplotype association statistics to [ " + f + " ]\n");
      HFRQ.open(f.c_str(), ios::out);
      HFRQ.precision(4);
      
      HFRQ << setw(10) << "LOCUS" << " " 
	   << setw(12) << "HAPLOTYPE" << " "
	   << setw(10) << "F_A" << " "
	   << setw(10) << "F_U" << " " 
	   << setw(10) << "CHISQ" << " "
	   << setw(4) << "DF" << " "
	   << setw(10) << "P" << "\n";
    }


  if (par::test_hap_TDT)
    {
      printLOG("Writing haplotype TDT statistics to [ " + f + " ]\n");
      HFRQ.open(f.c_str(), ios::out);
      HFRQ.precision(4);
      
      HFRQ << setw(10) << "LOCUS" << " " 
	   << setw(12) << "HAPLOTYPE" << " "
	   << setw(10) << "T" << " "
	   << setw(10) << "U" << " " 
	   << setw(10) << "CHISQ" << " "
	   << setw(10) << "P" << "\n";
    }
  




  //////////////////////////////
  // Begin phasing (and testing)
  
  int nms = new_map.size();
  for (int l=0; l<nms; l++)
    {
      
      if (!par::silent) 
	cout << l+1 << " out of " << nms << " haplotypes phased       \r";
      
      
      // Get number of predictors
      int s = new_pred_locus[l].size();
      
      // Same window as previous? 
      bool same = true;
      if (l==0) same = false;
      else
	{
	  if (new_pred_locus[l].size() != new_pred_locus[l-1].size() ) 
	    { same = false; }
	  else for (int j=0;j<s;j++)
	    {
	      if (new_pred_locus[l][j] != new_pred_locus[l-1][j]) 
		{ same = false; break; }
	    }
	}
            
            
      ///////////////////////////////////////////////////////////////
      // Calculate haplotype frequencies and posterior probabilities,
      // if different from previous round
      
      if (!same)  
	{
	  
	  ///////////////////////////////////////////////////
	  // Initialise
	  
	  H.reset();	
	  H.validN = cnt_f;

	  
	  // Is this a wildcard haplotype?
	  string nm = new_map[l]->name;
	  if (nm.substr(nm.size()-1) == "_") 
	    H.name(new_map[l]->name.substr(0,new_map[l]->name.find("_")));
	  else
	    H.name(new_map[l]->name);
	  

	  ///////////////////////////////////////////////////
	  // Generate binary list of all possible haplotypes
	  
	  H.enumerateHaplotypes(new_pred_locus[l]);


	  /////////////////////////////////////////////////
	  // Enumerate all possible phases for each founder
	  
	  for (int i=0; i<n; i++) 
	    if (sample[i]->founder)
	      H.enumeratePhase(i);
	  

	  ////////////////////////////////
	  // E-M phasing based on founders

	  H.performEM();


	  ////////////////////////
	  // Prune unlikely phases 

 	  for (int i=0; i<n; i++) 
 	    if (sample[i]->founder)
 	      H.prunePhase(i);


	  ///////////////////////////////////
	  // Fill-in phasing any non-founders
	  
	  if (H.nonfounders)
	  {

	    // List all possible non-rare phases
  	    H.enumerateAllPhases();

	    for (int i=0; i<n; i++) 
	      if (!sample[i]->founder)
		{
		  H.phaseNonfounder(i);
		  H.prunePhase(i);
		}
	  } 	      

	  /////////////////////////////////////////
	  // Verbose output: haplotypes/phases
	  
	  if (par::display_hap_freqs || par::test_hap_CC || par::test_hap_TDT ) 
	    H.reportHaplotypeFrequencies(HFRQ);
	  if (par::display_phase_probs) H.reportPhase(*this);

	}
      

      //////////////////////////////////////////
      // Set which is the test haplotype/allele
      
      H.setTestHaplotype(new_pred_allele[l]);
      

      //////////////////////////////////////////
      // Impute for all individiuals, if common haplotype
      
      double w = 0;
      double c = 0;
      
      if (H.testHaplotypeFreq() > par::min_af ) 
	{

	  for (int i=0; i<n; i++) 
	    {
	      bool b1, b2;	
	      
	      // Haplotype-weighting
	      double t = H.imputeHaplotypes(i, b1, b2);
	      if (t>=0)
		{
		  w += t;
		  c++;	  
		};
	      
	      
	      // And set new imputed genotypes
	      new_one[i].push_back(b1);
	      new_two[i].push_back(b2);
	      
	    }
	  

	  // TEMPORAILY REMOVE WEIGHTING FUNCTION
	  if (false)
	    WGT << new_map[l]->name << "\t"
		<< w/c << "\n";
	  
	}
      else
	{
	  // Remove rare haplotypes from map
	  new_map[l]->bp = -1;
	}
      
    }


  if (!par::silent) cout << "\n";

  
  if (false)
    WGT.close();
  
  HFRQ.close();


 if (par::impute_tags)
   {

     ///////////////////////////
     // Write new PED file

     f = par::output_file_name + ".impute.ped";
  
     printLOG("Writing imputed ped file to [ " + f + " ] \n");
     ofstream PED(f.c_str(), ios::out);
     PED.clear();

     string delimit = " ";
     if (par::recode_jes) delimit = "\t";
     
     for (int i=0;i<n;i++)
       {

	 Individual * person = sample[i];
	 PED << person->fid << " "
	     << person->iid << " "
	     << person->pat << " "
	     << person->mat << " "
	     << person->sexcode << " ";
	 
	 if (par::bt)
	   PED << (int)person->phenotype;
	 else
	   PED << person->phenotype;
	 
	 for (int l=0;l<new_one[i].size();l++)
	   {
	     if ( (!new_one[i][l]) && (!new_two[i][l]) )  
	       PED << delimit 
		   << new_map[l]->allele1 << " " 
		   << new_map[l]->allele1;
	     else if ( (!new_one[i][l]) && new_two[i][l]) 
	       PED << delimit 
		   << new_map[l]->allele1 << " " 
		   << new_map[l]->allele2;
	     else if (  new_one[i][l]   && new_two[i][l]) 
	       PED << delimit 
		   << new_map[l]->allele2 << " " 
		   << new_map[l]->allele2;
	     else PED << delimit << "0 0";
	   }
	 PED << "\n";
       }
     
     PED.close();
  
     
     //////////////////
     // Write new map
     
     f = par::output_file_name + ".impute.map";
     printLOG("Writing imputed map file to [ " + f + " ] \n");
     ofstream MAP(f.c_str(),ios::out);
     MAP.clear();
     for (int l=0; l<new_map.size(); l++)
       {
	 if (new_map[l]->bp >= 0)  // rare SNPs have been removed
	   MAP << new_map[l]->chr << "\t"
	       << new_map[l]->name << "\t"
	       << new_map[l]->pos << "\t"
	       << new_map[l]->bp  << "\n";
       }
     MAP.close();
   }

}


void HaploPhase::enumerateHaplotypes(vector<int> & s)
{  
  
  // Make list of haplotypes, and code as +/-
  
  S = s;
  ns = s.size();
  nh = (int)pow((double)2,ns);
  f.resize(nh);
  for (int i=0;i<nh;i++) f[i] = 1/(double)nh;

  ph_hap1.clear();
  ph_hap2.clear();
  ph_freq.clear();

  
  // Optionally, for haplotype-based TDT
  if (par::test_hap_TDT)
    {
      trans.clear();
      untrans.clear();

      trans.resize(nh,0);
      untrans.resize(nh,0);
    }
  
  unsigned int h=0;
  
  while(h<nh)
    {
      vector<bool> m1;
      
      unsigned int p=1;
      for (int s=0;s<ns;s++)
	{
	  if ( h & p ) m1.push_back(false);
	  else m1.push_back(true);
	  p <<= 1;
	}
      
      // Add to list of haplotypes
      hap.push_back(m1);
      
      // Add to HapMap
      hapmap.insert(make_pair(m1,h));
      
      // Consider next haplotype
      h++;     

   }

}

void HaploPhase::enumerateAllPhases()
{  

  // Make list of all non-rare phases

  for (int h1=0;h1<nh;h1++)
    for (int h2=h1;h2<nh;h2++)
      {
	   double freq = f[h1] * f[h2];
	   if (h1!=h2) freq *= 2;
	   if ( freq >= par::hap_min_phase_prob )
	     {
		    ph_freq.push_back( freq );
		    ph_hap1.push_back( h1 );
		    ph_hap2.push_back( h2 );

           }
      }  

  np = ph_hap1.size();

}



vector<string> HaploPhase::returnHaplotypes(vector<int> & slist)
{
  vector<string> str;
  
  enumerateHaplotypes(slist);
  
  for (int h=0; h<hap.size(); h++)     
    {
      string hstr;
      for (int s=0;s<ns;s++)
        if (!hap[h][s]) hstr += P.locus[S[s]]->allele1;
        else if (P.locus[S[s]]->allele2=="") hstr += "0";
	else hstr += P.locus[S[s]]->allele2;
      str.push_back(hstr);
    }
  return str;
}



void HaploPhase::setTestHaplotype(string t)
{

  // Create match template
  vector<bool> tmp(ns,false); 
  for (int s=0;s<ns;s++)
   if ( P.locus[S[s]]->allele1 == t.substr(s,1) )
     tmp[s] = true;

  // Consider each haplotype 

  test_hap = -1; 
  
  for (int h=0; h<hap.size(); h++)     
  {
    bool match = true;
    
    for (int s=0;s<ns;s++)
	{
	  if ( hap[h][s] != tmp[s] )
        {
          match = false;	  	  
          break;
        }
	}
    
    if (match) 
     {
      test_hap = h;
      break;
     }
  }
}


void HaploPhase::prunePhase(int i)
{
  
  // pp[i][z]
  // hap1[i][z]
  // hap2[i][z]
  
  if ( (!include[i]) || 
       (!ambig[i])   ) return;

  double psum = 0;
  
  vector<double> new_pp(0);
  vector<int> new_h1(0);
  vector<int> new_h2(0);

  for (int z=0; z < hap1[i].size(); z++)
    {
      if (pp[i][z] >= par::hap_min_phase_prob)
	{
	  new_pp.push_back( pp[i][z] );
	  psum += pp[i][z];
	  new_h1.push_back( hap1[i][z] );
	  new_h2.push_back( hap2[i][z] );
	}
    }
  
  // Normalise?
  if ( pp[i].size() > new_pp.size() )
    {
      for (int z=0; z < new_pp.size(); z++)
	new_pp[z] /= psum;
    }
  
  // Update
  pp[i] = new_pp;
  hap1[i] = new_h1;
  hap2[i] = new_h2;

}



void HaploPhase::enumeratePhase(int i)
{  

  vector<bool> s1(ns);
  vector<bool> s2(ns);
 
  // Flipping allele-coding for
  // homozygotes
  for (int s=0; s<ns; s++)
    {
      if (par::SNP_major)
	{	
	  s1[s] = P.SNP[S[s]]->one[i];
	  s2[s] = P.SNP[S[s]]->two[i];
	}
      else
	{
	  s1[s] = P.sample[i]->one[S[s]];
	  s2[s] = P.sample[i]->two[S[s]];
	}
      
      if ( s1[s] == s2[s] )
	{
	  s1[s] = !s1[s];
	  s2[s] = !s2[s];
	}
      
    }
  
  
  /////////////////////////////////////
  // Missing at any locus -> ambiguous
  // but too much missing -> exclude

  int nm = 0;
  for (int s=0; s<ns; s++)
    if ( s1[s] && !s2[s] )
      nm++;
	
  if (nm>0) ambig[i] = true;  
  if ((double)nm/(double)ns >= par::hap_missing_geno) 
    {
      include[i] = false;
      validN--;
      return;
    }


  ///////////////////////////////////////////////
  // 2 or more hets at any loci -> ambiguous
  
  int het=0;
  for (int s=0; s<ns; s++)
    if ( (!s1[s]) && s2[s] ) het++;
  if (het>1) ambig[i] = true;
  


  //////////////////////////////////////
  // Construct list of consistent phases
  
  if (!ambig[i])
    {		
      // Unambiguous means all no missing genotypes
      // and less than 2 hets
      
      // Match haplotype alleles
     	hap1[i].push_back( hapmap.find(s1)->second );
     	hap2[i].push_back( hapmap.find(s2)->second );
      
	
	}
  else
    {
      
      // For individuals with ambiguity 
      
      // Which are the ambiguous sites
      // (missing or heterozygous)

      vector<bool> het_site(ns,false);
      vector<bool> mis_site(ns,false);
      bool seenhet = false;

      int ambig_cnt=0;
      for (int s=0; s<ns; s++)
	{
	  // het
	  if ( (!s1[s]) && s2[s] )
	    {
	      if (seenhet)
		{
		  het_site[s] = true;
		  ambig_cnt++;
		}
	      else seenhet = true;
	    }
	  
	  // missing
	  if ( s1[s] && (!s2[s]) )
	    {
	      mis_site[s] = true;
	      ambig_cnt+=2;
	    }	  
	}


      int ambig_nh = (int)pow((double)2,ambig_cnt);

      vector<bool> h1(ns);
      vector<bool> h2(ns);
 
      int h=0;
      while(h<ambig_nh)
	{
	  vector<bool> m1;
	  
	  unsigned int p=1;
	  for (int s=0;s<ambig_cnt;s++)
	    {
	      if ( h & p ) m1.push_back(false);
	      else m1.push_back(true);
	      p <<= 1;
	    }

	  // Splice m1-variant into h1, h2 to 
	  // reconstruct ambiguous sites
	  int ac=0;
	  for (int s=0; s<ns; s++)
	    {
	      
	      // Reconstruct heterozygous sites
	      // (except the first one)

	      if (het_site[s])
		{
		  if (m1[ac]) 
		    {
		      h1[s] = true;
		      h2[s] = false;
		    }
		  else
		    {
		      h1[s] = false;
		      h2[s] = true;
		    }
		  ac++;
		}
	      else if (mis_site[s])
		{
		  // Make het
		  if (m1[ac]) 
		    {
		      if (m1[ac+1])
			{
			  h1[s] = false;
			  h2[s] = true;
			}
		      else
			{
			  h1[s] = true;
			  h2[s] = false;
			}
		    }
		  else 
		    {
		      // otherwise, select a homozygote
		      if (m1[ac+1])
			h1[s] = h2[s] = false;
		      else 
			h1[s] = h2[s] = true;
		    }

		  ac+=2;
		}
	      else
		{
		  // Maintain unambigous site
		  // (which might be 1st het)
		  h1[s] = s1[s];
		  h2[s] = s2[s];
		}
	    }
	  
	  
	  // Add to list of haplotypes
	  int n1 = hapmap.find( h1 )->second;
	  int n2 = hapmap.find( h2 )->second;
	  
	  // Add? (always add...)
//	  if ( n1 <= n2 )
//	    {
	      hap1[i].push_back( n1 );
	      hap2[i].push_back( n2 );	      
//	    }

	  // Consider next 'haplotype'
	  h++;     
	  
	}
    }
  
 
  // Make space for posterior probabilities
  
  if (ambig[i]) pp[i].resize(hap1[i].size());


}

void HaploPhase::reportPhase(Plink & P)
{
   string fn = par::output_file_name+".phase-"+hname;
   ofstream PHASE(fn.c_str(), ios::out);

   P.printLOG("Writing phased haplotypes for " + hname + " to [ " + fn + " ]\n");
  
   PHASE << setw(12) << "FID" << " "
  	   << setw(12) << "IID" << " "
	   << setw(4)  << "PH" << " "
         << setw(10) << "HAP1" << " "
	   << setw(10) << "HAP2" << " "
	   << setw(12) << "POSTPROB" << " "
     //  << setw(12) << "WEIGHT" << " "
         << setw(6) << "BEST" << " "
	   << "\n";

  PHASE.precision(4);

  for (int i = 0 ; i < P.n ; i++ ) 
    {
      
      if (include[i])
	{
	  
	  for (int z = 0 ; z < hap1[i].size(); z++) 
	    {
	      
	      PHASE << setw(12) << P.sample[i]->fid << " " 
		    << setw(12) << P.sample[i]->iid << " "
		    << setw(4) << z << " "
		    << setw(10) << haplotypeName(hap1[i][z])  << " "
		    << setw(10) << haplotypeName(hap2[i][z]) << " ";
	      
	      if (ambig[i])	
		{	
		  PHASE << setw(12) << pp[i][z] << " ";
		  int max_z = 0;
		  for (int z2=0; z2<hap1[i].size(); z2++)
		    max_z = pp[i][z2] > pp[i][max_z] ? z2 : max_z ;
		  
		  
		  int fac=1;
		  if ( hap1[i][max_z] != hap2[i][max_z] ) fac=2;
		  
		  double w = ( pp[i][max_z] - fac*f[hap1[i][max_z]]*f[hap1[i][max_z]] ) 
		    / ( 1 - fac*f[hap1[i][max_z]]*f[hap1[i][max_z]] );
		  
		  // Do not output weight for now
		  // 	    if (max_z == z) PHASE << setw(12) << w  << " " << setw(6) << 1  << " " << "\n";
		  // 	    else PHASE << setw(12) << "."  << " " << setw(6) << 0  << " " << "\n";
		  
		  if (max_z == z) PHASE << setw(6) << 1  << " " << "  ";
		  else PHASE << setw(6) << 0  << " " << "  ";
		}
	      else
		// 	  PHASE << setw(12) << 1  << " "
		// 		<< setw(12) << 1  << " "
		// 		<< setw(6) << 1  << " "
		// 		<< "\n"; 
		PHASE << setw(12) << 1  << " "
		      << setw(6) << 1  << " "
		      << "  "; 

	// genotypes
       for (int s=0; s<ns; s++)
         PHASE << genotype(P,i,S[s]) << " ";
       PHASE << "\n";


	    }
	}

      // Report also on excluded individuals
      // (Should be 0-size phase-set)
      else 
	{

	  PHASE << setw(12) << P.sample[i]->fid << " " 
		<< setw(12) << P.sample[i]->iid << " "
		<< setw(4) << "NA" << " "
		<< setw(10) << "NA"  << " "
		<< setw(10) << "NA" << " "
		<< setw(12) << "NA"  << " "
		<< setw(6) << "NA"  << "  ";

	// genotypes
       for (int s=0; s<ns; s++)
         PHASE << genotype(P,i,S[s]) << " ";
       PHASE << "\n";


	}
    }

  PHASE.close();
  
}  


void HaploPhase::performEM()
{
  
  vector<double> uc(nh,0);  // unambigous counts
  vector<double> ac(nh,0);  // ambigous counts
  
  // Count numbers of unambigous haplotypes
  // as these stay constant throughout EM

  for (int i=0; i<P.n; i++)
    {
      if (P.sample[i]->founder && include[i])
	{
	  if (!ambig[i])
	    {
	      uc[hap1[i][0]]++;
	      uc[hap2[i][0]]++;
	    }
	}
    }
  
  
  //////////////////  
  // Begin E-M

  // TODO: Fixed number of iterations
  // We can really do better than this...
  // Implement to consider lnL
  // Also to drop zero frequency haplotypes?

  // But -- for 2-SNP analysis, neither of these
  // things really matter.
  
  for (int j=0; j<=20; j++)
    {
      
      ///////////
      // E-step
      
      for (int i=0; i<P.n; i++)
	{
	  if (P.sample[i]->founder && include[i])
	    {
	      if (ambig[i])
		{
		  double s=0;
		  for (int z=0; z<hap1[i].size(); z++)
		    {
		      pp[i][z] = f[hap1[i][z]] * f[hap2[i][z]];
		      if (hap1[i][z] != hap2[i][z]) pp[i][z] *= 2;
		      s += pp[i][z];
		    }
		  
		  for (int z=0; z<hap1[i].size(); z++) pp[i][z] /= s;
		}
	    }
	}
      


      ///////////
      // M-step
      
      // unambigous counts
      for (int h=0; h<nh; h++)
	f[h] = uc[h];
      
      // then add the fractional ones
      for (int i=0; i<P.n; i++)
	if (P.sample[i]->founder && include[i])
	  if (ambig[i])
	    {
	      for (int z=0; z<hap1[i].size(); z++)
		{
		  f[hap1[i][z]] += pp[i][z];
		  f[hap2[i][z]] += pp[i][z];		  
		}
	    }
      
      for (int h=0; h<nh; h++)      
	f[h] /= (double)validN*2;
      
    }
  
}


void HaploPhase::reportHaplotypeFrequencies(ofstream & HFRQ)
{


  //////////////////////////////////////////
  // Option to report haplotype frequencies

  if (par::display_hap_freqs)
    {
      
      for (int h=0; h<nh; h++)      
	{
	  if (f[h] >= par::min_af)
	    {
	      HFRQ << setw(10) << hname << " "
		   << setw(12) << haplotypeName(h) << " " 
		   << setw(10) << f[h] << "\n";
	    }
	}
      
      
      return;
    }
  


  ////////////////////////////////////////////////
  // Perform omnibus and haplotype-specific tests

  if (! ( par::test_hap_CC || par::test_hap_TDT ) ) 
    return;

  // Omnibus test

  // tests[0] = 0
  // tests[1] = 1
  // tests[2] = 2
  // ...
  // tests[h] = h
  
  map<int,int> tests;
  int nch=0;
  for (int h=0; h < nh; h++)
    if (f[h] >= par::min_af)
      {
	tests.insert(make_pair(h,nch++));
      }

  if (nch>2)
    {
      if (par::test_hap_CC)
	haplotypicCC(tests,nch,HFRQ);
      else if (par::test_hap_TDT)
	haplotypicTDT(tests,nch,HFRQ);
    }

  // Haplotype-specific test

  // tests[0] = 0
  // tests[1] = 1,2,...,h

  // tests[0] = 1
  // tests[1] = 0,2,3,...,h

  // etc

  for (int h=0; h < nh; h++)
    {
      if (f[h] >= par::min_af)
	{
	  tests.clear();
	  
	  for (int h2=0; h2 < nh; h2++)
	    {
	      if (f[h2] >= par::min_af)
		{
		  if (h==h2) 
		    {
		      tests.insert(make_pair(h2,0));
		    }
		  else tests.insert(make_pair(h2,1));
		}
	    }
	  
	  if (par::test_hap_CC)
	    haplotypicCC(tests,2,HFRQ);
	  else if (par::test_hap_TDT)
	    haplotypicTDT(tests,2,HFRQ);
	}
    }
 
   
} 

 

double HaploPhase::imputeHaplotypes(int i, bool & n1, bool & n2)
{

  //////////////////////////////////////////////
  // Based on P(H|G) impute inferred haplotypes
  // for above-threshold individuals  

  // Not imputed
  double w = -1;
  
  // First for individuals of unambiguous phase

  if (include[i])
    {    
      if (!ambig[i])
	{
	  if (hap1[i][0] == test_hap) n1 = false;
	  else n1 = true;
	  
	  if (hap2[i][0] == test_hap) n2 = false;
	  else n2 = true;
	  
	  // Resolve potential het/missing coding confusion
	  
	  if ( n1 && (!n2) )
	    {
	      n1 = false;
	      n2 = true;	      
	    }
	  
	  // Unambiguous weighting (0, 1 or 2 copie of test_hap)
	  if (!n1)
	    {
	      if (!n2) w = 2;
	      else w = 1;
	    }
	  else w = 0;
	}
      else
	{
	  
	  // Second, for ambiguous individuals impute and assign weight
	  
	  int max_z = 0;
	  for (int z=0; z<hap1[i].size(); z++)
	    max_z = pp[i][z] > pp[i][max_z] ? z : max_z ;
	  
	  // Set missing by default
	  
	  n1 = true;
	  n2 = false;	      
	  
	  // Consider each phase z
	  
	  // Above threshold? 
	  if (pp[i][max_z] >= par::hap_post_prob)
	    {
	      
	      // Do we match 'test_hap' ( '1' allele ) 
	      // or not? ( '2' allele )
	      
	      if (hap1[i][max_z] == test_hap) n1 = false;
	      else n1 = true;
	  
	      if (hap2[i][max_z] == test_hap) n2 = false;
	      else n2 = true;
	  
	      // Resolve potential het/missing coding confusion
	  
	      if ( n1 && (!n2) )
		{
		  n1 = false;
		  n2 = true;	      
		}		  
	      
	  
	      // Unambiguous weighting (1 or 2 copie of test_hap)

	      // We are saying either 0, 1 or 2 copies
	      // Consider each haplotype

	      // Number imputed / Actual number
	      for (int z=0; z<hap1[i].size(); z++)
		{
		  if (hap1[i][z] == test_hap) w += pp[i][z];
		  if (hap2[i][z] == test_hap) w += pp[i][z];
		}
	  
	      w = pp[i][max_z] / w;

	      // 	  if (!n1)
	      //  	    {
	      //  	      if (!n2) w = pp[i][test_hap] * 1;
	      //  	      else w = pp[i][test_hap] * 0.5;
	      //  	    }
	      
	      // 	  int fac=1;
	      // 	  if ( hap1[i][max_z] != hap2[i][max_z] ) fac=2;
	      // 	  w = ( pp[i][max_z] - fac*f[hap1[i][max_z]]*f[hap1[i][max_z]] ) / ( 1 - fac*f[hap1[i][max_z]]*f[hap1[i][max_z]] );
	      
	    }
	}     
    }
  else
    {
      // Excluded inviduals
      n1 = true;
      n2 = false;
    }
  
  
  return w;
}



void HaploPhase::phaseNonfounder(int i)
{

  //////////////////////////////////////////////
  // Link this individual up with their parents

  int father = P.sample[i]->ip;
  int mother = P.sample[i]->im;
  
  bool nofather = false;
  bool nomother = false;

  if (father==-1) 
    nofather = true;
  else if (!include[father])
    nofather = true;

  if (mother==-1) 
    nomother = true;
  else if (!include[mother])
    nomother = true;


  // For TDT purposes, we require both parents to be 'observed'

  if (nofather || nomother) 
  {
     include[i] = false;
     return;
  }

  int pat_phases = hap1[father].size();
  int mat_phases = hap1[mother].size();


  // Too much ambiguity?

  if ( pat_phases * mat_phases >= par::hap_max_nf_phases )
    {
      include[i] = false;
      return;
    }


  // Keep track of transmitted and non-transmitted 
  // haplotypes if performing a TDT-type analysis

  vector<vector<int> > trans1(0);
  vector<vector<int> > untrans1(0);
  

  /////////////////////////////////////////
  // Perform fill-in phasing for offspring
  
  // Step 1. Enumerate possible offspring phases
  //         or set to not include if too much missing

  enumeratePhase(i);
  
  // Do we want to attempt to reconstruct phase?
  
  if ( !include[i] ) 
  { 
      include[i] = false;
      return;
  }

  // Step 2. Joint distribution of parental phases
  
  vector<double> prob(0);
  double psum = 0;
  
  for (int z=0; z < hap1[i].size() ; z++)
    {
      
      // Is this phase possible given parental haplotypes?
      int h1 = hap1[i][z];
      int h2 = hap2[i][z];
      
      for (int z1=0; z1 < pat_phases ; z1++)
	for (int z2=0; z2 < mat_phases ; z2++)
	  {	
	    
	    // If no mother or father exists, we are 
	    // using the standard ph[] enumeration of
	    // all possible haplotypes
	    
	    int p1, p2, m1, m2;
	    
	    if (nofather)
	      {
		p1 = ph_hap1[z1];
		p2 = ph_hap2[z1];
	      }
	    else
	      {
		p1 = hap1[father][z1];
		p2 = hap2[father][z1];
	      }
	    	    
	    if (nomother)
	      {
		m1 = ph_hap1[z2];
		m2 = ph_hap2[z2];
	      }
	    else
	      {
		m1 = hap1[mother][z2];
		m2 = hap2[mother][z2];
	      }
	    
	    if ( ( h1 == p1 && h2 == m1 ) || 
		 ( h1 == p1 && h2 == m2 ) || 
		 ( h1 == p2 && h2 == m1 ) || 
		 ( h1 == p2 && h2 == m2 ) || 
		 ( h1 == m1 && h2 == p1 ) || 
		 ( h1 == m1 && h2 == p2 ) || 
		 ( h1 == m2 && h2 == p1 ) || 
		 ( h1 == m2 && h2 == p2 ) )
	      {		  

		double p = 1;
		
		if (nofather)
		  p *= ph_freq[z1];
		else
		  if (ambig[father]) p *= pp[father][z1];
		
		if (nomother)
		  p *= ph_freq[z2];
		else
		  if (ambig[mother]) p *= pp[mother][z2];
		
		if (h1!=h2) p *= 2;
		
		psum += p;
		

// Consider -- is this possibility worth considering???
// + is this the right place to consider it????
		prob.push_back(p);
		
		if (ambig[i])
		  pp[i][z] += p;
		
		if (par::test_hap_TDT)
		  {
		    		    
		    vector<int> t1(nh);
		    vector<int> u1(nh);
		    
		    // Father heterozygous? 
		    if ( p1 != p2 )
		      {
			// Mother homozygous?
			if ( m1 == m2 ) 
			  {
			    // then select a kid allele that matches, 
			    // and score the other one for transmission
			    if ( h1 == m1 ) 
			      {
				t1[h2]++;
				if (p1==h2)
				  u1[p2]++;
				else
				  u1[p1]++;
			      }
			    else
			      {
				t1[h1]++;
				if (p1==h1)
				  u1[p2]++;
				else
				  u1[p1]++;
				
			      }
			  }
			else
			  {
			    // Both parents are heterozygous, 
			    
			    // Transmitted alleles are unambiguous
			    t1[h1]++;
			    t1[h2]++;
			    
			    // Untransmitted alleles
			    // i.e. which two are left over 
			    // after accounting for the two
			    // transmitted alleles
			    
			    bool pat_accounted = false;
			    bool mat_accounted = false;
			    if ( p1 != h1 && p1 != h2 )
			      {
				u1[p1]++;
				pat_accounted = true;
			      }
			    else if ( p2 != h1 && p2 != h2 )
			      {
				u1[p2]++;
				pat_accounted = true;
			      }
			    
			    if ( m1 != h1 && m1 != h2 )
			      {
				u1[m1]++;
				mat_accounted = true;
			      }
			    else if  ( m2 != h1 && m2 != h2 )
			      {
				u1[m2]++;
				mat_accounted = true;
			      }
			    
			    // This only happens with AB x AB -> AB
			    if ( ! ( pat_accounted || mat_accounted ) ) 
			      {
				u1[h1]++;
				u1[h2]++;				    
			      }
			    else if ( ( ( !pat_accounted ) &&   mat_accounted   ) || 
				      (    pat_accounted   && (!mat_accounted)  ) )
			      {
				// If only 1 untransmitted allele accounted, for, it must
				// be the doubled allele that is untransmitted)
				
				if ( p1 == m1 || p1 == m2 ) 
				  u1[p1]++;
				else
				  u1[p2]++;
			      }
			    
			    
			    // AB AB    BB -- 2 acconuted for
			    //  AA
			    
			    // AB AB    AB -- 0 accounted for
			    //  AB 
			    
			    // AB AC    BA -- 1 accounted for
			    //  AC
			    
			    // AC AB    BA -- 1 accounted for
			    //  AC
			    
			    // AB CD    DB  -- 2 accounted for 
			    //  AC
			    
			  }
		      }
		    else if ( m1 != m2 ) 
		      {
			// Mother heterozygous, father homozygous
			if ( h1 == p1 )
			  {
			    t1[h2]++;
			    if (m1==h2)
			      u1[m2]++;
			    else
			      u1[m1]++;
			  }
			else
			  {
			    t1[h1]++;
			    if (m1==h1)
			      u1[m2]++;
			    else
			      u1[m1]++;
			  }
		      }
		    
		    
		    trans1.push_back(t1);
		    untrans1.push_back(u1);			
		    
		    
		    if (par::verbose)
		      {
			cout << P.sample[i]->fid << " "
			     << P.sample[i]->iid << "\t";
			
		      }

		  }
		
		
	      }
	  } 
    }
  
  
  // Normalise probabilities
  for (int z=0; z < prob.size(); z++)
    prob[z] /= psum;
  
  if (ambig[i])
    for (int z=0; z < pp[i].size(); z++)
      pp[i][z] /= psum;

  
  // Finished counting -- add to sample trans/untrans counts
  if (par::test_hap_TDT)
    {      
      for (int h=0; h<nh; h++)
	{
	  for (int z=0; z<prob.size(); z++)
	    {
	      trans[h] += trans1[z][h] * prob[z];
	      untrans[h] += untrans1[z][h] * prob[z];
	    }
	}
    }

  return; 
}
